/**
 * @file Layout.cpp
 * @author apio (cloudapio.eu)
 * @brief Layout widgets to organize content.
 *
 * @copyright Copyright (c) 2023, the Luna authors.
 *
 */

#include <stdlib.h>
#include <ui/Layout.h>
#include <ui/Window.h>

namespace ui
{
    HorizontalLayout::HorizontalLayout(Window* window, Widget* parent) : ui::Widget(window, parent)
    {
    }

    Result<EventResult> HorizontalLayout::handle_mouse_move(Point position)
    {
        EventResult result = ui::EventResult::DidNotHandle;

        for (auto widget : m_widgets)
        {
            if (widget->rect().contains(position)) result = TRY(widget->handle_mouse_move(position));
            else
                TRY(widget->handle_mouse_leave());
        }

        return result;
    }

    Result<EventResult> HorizontalLayout::handle_mouse_leave()
    {
        EventResult result = ui::EventResult::DidNotHandle;

        for (auto widget : m_widgets)
        {
            auto rc = TRY(widget->handle_mouse_leave());
            if (rc == ui::EventResult::DidHandle) result = rc;
        }

        return result;
    }

    Result<EventResult> HorizontalLayout::handle_mouse_up(Point position, int buttons)
    {
        for (auto widget : m_widgets)
        {
            if (widget->rect().contains(position)) return widget->handle_mouse_up(position, buttons);
        }

        return ui::EventResult::DidNotHandle;
    }

    Result<EventResult> HorizontalLayout::handle_mouse_down(Point position, int buttons)
    {
        for (auto widget : m_widgets)
        {
            if (widget->rect().contains(position)) return widget->handle_mouse_down(position, buttons);
        }

        return ui::EventResult::DidNotHandle;
    }

    Result<EventResult> HorizontalLayout::handle_key_event(const ui::KeyEventRequest& request)
    {
        EventResult result = ui::EventResult::DidNotHandle;

        for (auto widget : m_widgets)
        {
            auto rc = TRY(widget->handle_key_event(request));
            if (rc == ui::EventResult::DidHandle) result = rc;
        }

        return result;
    }

    Result<void> HorizontalLayout::draw(Canvas& canvas)
    {
        for (auto widget : m_widgets)
        {
            ui::Rect rect = { m_rect.relative(widget->rect().pos), widget->rect().width, widget->rect().height };
            auto subcanvas = canvas.subcanvas(rect);
            TRY(widget->draw(subcanvas));
        }

        return {};
    }

    Result<void> HorizontalLayout::add_widget(Widget& widget)
    {
        TRY(m_widgets.try_append(&widget));

        window()->recalculate_layout();

        return {};
    }

    void HorizontalLayout::recalculate_pseudo_rects()
    {
        int min_height = 0;
        int width = 0;

        for (auto w : m_widgets)
        {
            w->recalculate_pseudo_rects();
            if (w->pseudo_rect().height > min_height) min_height = w->pseudo_rect().height;
            width += (m_margins.left + w->pseudo_rect().width + m_margins.right);
        }

        min_height += (m_margins.top + m_margins.bottom);

        m_pseudo_rect = { 0, 0, width, min_height };

        if (m_preferred_rect.has_value()) m_pseudo_rect = m_preferred_rect.value();
    }

    void HorizontalLayout::recalculate_true_rects()
    {
        if (m_adjust_width == AdjustWidth::Yes)
        {
            int used_width = 0;
            div_t result = div(m_rect.width, (int)m_widgets.size());
            for (auto w : m_widgets)
            {
                w->rect().pos.x = m_rect.pos.x + used_width + m_margins.left;
                w->rect().pos.y = m_rect.pos.y + m_margins.top;
                w->rect().width = result.quot - (m_margins.left + m_margins.right);

                if (m_adjust_height == AdjustHeight::Yes)
                {
                    w->rect().height = m_rect.height - (m_margins.top + m_margins.bottom);
                }
                else
                    w->rect().height = w->pseudo_rect().height;

                used_width += result.quot;

                w->recalculate_true_rects();
            }

            if (result.rem != 0 && m_widgets.size())
            {
                auto* last = m_widgets[m_widgets.size() - 1];
                last->rect().width += result.rem;
                last->recalculate_true_rects();
            }
        }
        else
        {
            int used_width = 0;
            for (auto w : m_widgets)
            {
                w->rect() = w->pseudo_rect();
                w->rect().pos.x = m_rect.pos.x + used_width + m_margins.left;
                w->rect().pos.y = m_rect.pos.y + m_margins.top;

                if (m_adjust_height == AdjustHeight::Yes)
                {
                    w->rect().height = m_rect.height - (m_margins.top + m_margins.bottom);
                }
                else
                    w->rect().height = w->pseudo_rect().height;

                used_width += (m_margins.left + m_margins.right + w->rect().width);

                w->recalculate_true_rects();
            }
        }
    }

    VerticalLayout::VerticalLayout(Window* window, Widget* parent) : ui::Widget(window, parent)
    {
    }

    Result<EventResult> VerticalLayout::handle_mouse_move(Point position)
    {
        EventResult result = ui::EventResult::DidNotHandle;

        for (auto widget : m_widgets)
        {
            if (widget->rect().contains(position)) result = TRY(widget->handle_mouse_move(position));
            else
                TRY(widget->handle_mouse_leave());
        }

        return result;
    }

    Result<EventResult> VerticalLayout::handle_mouse_leave()
    {
        EventResult result = ui::EventResult::DidNotHandle;

        for (auto widget : m_widgets)
        {
            auto rc = TRY(widget->handle_mouse_leave());
            if (rc == ui::EventResult::DidHandle) result = rc;
        }

        return result;
    }

    Result<EventResult> VerticalLayout::handle_mouse_up(Point position, int buttons)
    {
        for (auto widget : m_widgets)
        {
            if (widget->rect().contains(position)) return widget->handle_mouse_up(position, buttons);
        }

        return ui::EventResult::DidNotHandle;
    }

    Result<EventResult> VerticalLayout::handle_mouse_down(Point position, int buttons)
    {
        for (auto widget : m_widgets)
        {
            if (widget->rect().contains(position)) return widget->handle_mouse_down(position, buttons);
        }

        return ui::EventResult::DidNotHandle;
    }

    Result<EventResult> VerticalLayout::handle_key_event(const ui::KeyEventRequest& request)
    {
        EventResult result = ui::EventResult::DidNotHandle;

        for (auto widget : m_widgets)
        {
            auto rc = TRY(widget->handle_key_event(request));
            if (rc == ui::EventResult::DidHandle) result = rc;
        }

        return result;
    }

    Result<void> VerticalLayout::draw(Canvas& canvas)
    {
        for (auto widget : m_widgets)
        {
            ui::Rect rect = { m_rect.relative(widget->rect().pos), widget->rect().width, widget->rect().height };
            auto subcanvas = canvas.subcanvas(rect);
            TRY(widget->draw(subcanvas));
        }

        return {};
    }

    void VerticalLayout::recalculate_pseudo_rects()
    {
        int min_width = 0;
        int height = 0;

        for (auto w : m_widgets)
        {
            w->recalculate_pseudo_rects();
            if (w->pseudo_rect().width > min_width) min_width = w->pseudo_rect().width;
            height += (m_margins.top + w->pseudo_rect().height + m_margins.bottom);
        }

        min_width += (m_margins.left + m_margins.right);

        m_pseudo_rect = { 0, 0, min_width, height };

        if (m_preferred_rect.has_value()) m_pseudo_rect = m_preferred_rect.value();
    }

    void VerticalLayout::recalculate_true_rects()
    {
        if (m_adjust_height == AdjustHeight::Yes)
        {
            int used_height = 0;
            div_t result = div(m_rect.height, (int)m_widgets.size());
            for (auto w : m_widgets)
            {
                w->rect().pos.y = m_rect.pos.y + used_height + m_margins.top;
                w->rect().pos.x = m_rect.pos.x + m_margins.left;
                w->rect().height = result.quot - (m_margins.top + m_margins.bottom);

                if (m_adjust_width == AdjustWidth::Yes)
                {
                    w->rect().width = m_rect.width - (m_margins.left + m_margins.right);
                }
                else
                    w->rect().width = w->pseudo_rect().width;

                used_height += result.quot;

                w->recalculate_true_rects();
            }

            if (result.rem != 0 && m_widgets.size())
            {
                auto* last = m_widgets[m_widgets.size() - 1];
                last->rect().height += result.rem;
                last->recalculate_true_rects();
            }
        }
        else
        {
            int used_height = 0;
            for (auto w : m_widgets)
            {
                w->rect() = w->pseudo_rect();
                w->rect().pos.y = m_rect.pos.y + used_height + m_margins.top;
                w->rect().pos.x = m_rect.pos.x + m_margins.left;

                if (m_adjust_width == AdjustWidth::Yes)
                {
                    w->rect().width = m_rect.width - (m_margins.left + m_margins.right);
                }
                else
                    w->rect().width = w->pseudo_rect().width;

                used_height += (m_margins.top + m_margins.bottom + w->rect().height);

                w->recalculate_true_rects();
            }
        }
    }

    Result<void> VerticalLayout::add_widget(Widget& widget)
    {
        TRY(m_widgets.try_append(&widget));

        window()->recalculate_layout();

        return {};
    }
}
